from typing import Any, Optional, ClassVar
from dataclasses import dataclass
from http import HTTPStatus
import traceback

from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.exceptions import HTTPException

@dataclass
class HTTPError(HTTPException):
	desc: str
	data: Optional[Any] = None
	code: ClassVar[int]

	def __post_init__(self) -> None:
		super().__init__(self.code, self.desc)

	def as_response(self) -> JSONResponse:
		return error_response(self.code, self.desc, self.data)

@dataclass
class BadRequest(HTTPError):
	code = 400

@dataclass
class Unauthorised(HTTPError):
	code = 401

@dataclass
class Forbidden(HTTPError):
	code = 403

@dataclass
class NotFound(HTTPError):
	code = 404

@dataclass
class Conflict(HTTPError):
	code = 409

@dataclass
class TooManyRequests(HTTPError):
	code = 429

@dataclass
class InternalServerError(HTTPError):
	code = 500

def error_response(code: int, desc: Optional[str], data: Optional[Any]) -> JSONResponse:
	return JSONResponse({
		'error_type': HTTPStatus(code).phrase,
		'error_description': desc,
		'error_data': data,
	}, code)

class JSONErrorMiddleware(BaseHTTPMiddleware):
	async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
		# TODO: logging
		try:
			response = await call_next(request)
			return response
		except HTTPError as e:
			return e.as_response()
		except Exception as e:
			traceback.print_exc()
			return error_response(500, None, None)



from typing import Optional, Any, Dict, Union, Awaitable, Callable, Sequence, Iterator
from dataclasses import dataclass
from time import perf_counter
from contextlib import contextmanager

from starlette.types import ASGIApp
from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
import aioprometheus

Observation = Union[float, int]
LabelValue = Union[str, int]

@dataclass
class LabelUpdater:
	_labels: Dict[str, LabelValue]

	def update(self, **labels: LabelValue) -> None:
		self._labels = { **self._labels, **labels }

@dataclass
class Counter:
	_counter: aioprometheus.Counter

	def inc(self, **labels: LabelValue) -> None:
		self._counter.inc(labels)

	def add(self, val: int, **labels: LabelValue) -> None:
		self._counter.add(labels, val)

@dataclass
class Gauge:
	_gauge: aioprometheus.Gauge

	def set(self, val: Observation, **labels: LabelValue) -> None:
		self._gauge.set(labels, val)

	def inc(self, **labels: LabelValue) -> None:
		self._gauge.inc(labels)

@dataclass
class Histogram:
	_histogram: aioprometheus.Histogram

	def observe(self, obs: Observation, **labels: LabelValue) -> None:
		self._histogram.observe(labels, obs)

	@contextmanager
	def observe_time(self, **labels: LabelValue) -> Iterator[LabelUpdater]:
		start_time = perf_counter()
		label_updater = LabelUpdater(labels)
		yield label_updater
		elapsed_secs = perf_counter() - start_time
		self.observe(elapsed_secs, **label_updater._labels)

class MetricsRegistry:
	def __init__(self) -> None:
		self._prom_svc = aioprometheus.Service()
		self._const_labels: Dict[str, LabelValue] = {}
	
	def counter(self, name: str, desc: str = '') -> Counter:
		counter = aioprometheus.Counter(name, desc, self._const_labels)
		ret = Counter(counter)
		self._prom_svc.register(counter)
		return ret

	def histogram(self, name: str, desc: str = '', buckets: Sequence[float] = aioprometheus.Histogram.DEFAULT_BUCKETS) -> Histogram:
		histogram = aioprometheus.Histogram(name, desc, self._const_labels, buckets = buckets)
		ret = Histogram(histogram)
		self._prom_svc.register(histogram)
		return ret

	def gauge(self, name: str, desc: str = '') -> Gauge:
		gauge = aioprometheus.Gauge(name, desc, self._const_labels)
		ret = Gauge(gauge)
		self._prom_svc.register(gauge)
		return ret

class MetricsMiddleware(BaseHTTPMiddleware):
	def __init__(self, app: ASGIApp, metrics_registry: MetricsRegistry) -> None:
		super().__init__(app)
		self.request_timing_histogram = metrics_registry.histogram('api_response_time_seconds')

	async def dispatch(self, req: Request, call_next: RequestResponseEndpoint) -> Response:
		with self.request_timing_histogram.observe_time(method = req.method, path = req.url.path) as labels:
			res = await call_next(req)
			labels.update(status = res.status_code)

		return res



from typing import List
from dataclasses import dataclass, field

from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

@dataclass
class CORSConfig:
	origins: List[str]
	methods: List[str]
	request_headers: List[str]
	response_headers: List[str]
	include_cookies: bool
	max_age_secs: int

def permissive_cors_config() -> CORSConfig:
	return CORSConfig(
		origins = [ '*' ],
		methods = [ '*' ],
		request_headers = [ '*' ],
		response_headers = [ '*' ],
		include_cookies = False,
		max_age_secs = 600,
	)

def cors_middleware_from_config(config: CORSConfig) -> Middleware:
	return Middleware(
		CORSMiddleware,
		allow_origins = config.origins,
		allow_methods = config.methods,
		allow_headers = config.request_headers,
		allow_credentials = config.include_cookies,
		expose_headers = config.response_headers,
	)



from typing import Dict, Optional, Any

import uvicorn

from .app import create_app
from .auth import AuthBackend
from .security import CORSConfig

def run(
		srv: Any,
		name: Optional[str] = None,
		auth_backend: Optional[AuthBackend] = None,
		cors_config: Optional[CORSConfig] = None,
		port: int = 80,
		metrics_port: int = 5000,
		host: str = "0.0.0.0",
		uvicorn_kwargs: Dict[str, Any] = {},
	) -> None:
	app = create_app(srv, name, auth_backend, cors_config, metrics_port)
	uvicorn.run(app, host = host, port = port, **uvicorn_kwargs)



from collections import defaultdict
from dataclasses import fields, is_dataclass
from datetime import date, datetime
from enum import Enum, EnumMeta
from typing import Any, Dict, List, Literal, Optional, Union

from .auth import AuthBackend
from .route_def import RouteDef

_pytype_to_schema_type_lookup = {
	int: 'integer',
	str: 'string',
	float: 'number',
	bool: 'boolean',
	type(None): 'null',
	datetime: 'string',
	date: 'string'
}

_pytype_string_formats = {
	datetime: 'date-time',
	date: 'date',
}

def pytype_to_schema(py_type: type) -> Dict[str, Any]:
	origin_type = getattr(py_type, '__origin__', None)
	generic_types = getattr(py_type, '__args__', None)

	if not origin_type:
		if is_dataclass(py_type):
			return {
				'type': 'object',
				'properties': { f.name: pytype_to_schema(f.type) for f in fields(py_type) },
			}
		elif isinstance(py_type, EnumMeta):
			return {
				'type': 'string',
				'enum': list(e.value for e in py_type) # type: ignore
			}
	else:
		if origin_type == list:
			return {
				'type': 'array',
				'items': pytype_to_schema(generic_types[0]),
			}
		elif origin_type == dict:
			if generic_types[0] is not str:
				raise Exception('dictionary keys must be strings')

			return {
				'type': 'object',
				'additionalProperties': pytype_to_schema(generic_types[1]),
			}
		elif origin_type == Union:
			return {
				'oneOf': [ pytype_to_schema(t) for t in generic_types ],
			}
		elif origin_type == Literal:
			return {
				'type': _pytype_to_schema_type_lookup[type(generic_types[0])],
				'enum': generic_types,
			}

	return {
		'type': _pytype_to_schema_type_lookup[py_type],
		'format': _pytype_string_formats.get(py_type),
	}

def build_schemas(route_defs: List[RouteDef], api_name: str, auth_backend: Optional[AuthBackend]) -> Dict[str, Any]:
	paths: Dict[str, Any] = defaultdict(dict)
	for route_def in route_defs:
		spec: Any = {
			'summary': route_def.readable_name,
			'description': route_def.doc,
			'responses': {
				'200': {
					'description': 'success',
					'content': {
						'application/json': {
							'schema': pytype_to_schema(route_def.return_type),
						},
					},
				},
			}
		}

		if route_def.http_method == 'get':
			spec['parameters'] = [ {
					'name': arg_name,
					'in': 'query',
					'schema': pytype_to_schema(arg_def.original_type),
					'required': not arg_def.optional,
				}
				for arg_name, arg_def in route_def.args.items()
			]
		else:
			spec['requestBody'] = {
				'required': True,
				'content': {
					'application/json': {
						'schema': pytype_to_schema(list(route_def.args.values())[0].original_type),
					},
				},
			}

		paths[route_def.path][route_def.http_method] = spec

	openapi_spec: Dict[str, Any] = {
		'openapi': '3.0.0',
		'info': {
			'title': api_name,
			'version': '0.1',
		},
		'paths': paths,
		'components': {},
	}

	if auth_backend:
		auth_name, spec = auth_backend.openapi_spec()
		openapi_spec['components']['securitySchemes'] = {
			auth_name: spec,
		}
		openapi_spec['security'] = [
			{ auth_name: [] },
		]

	return openapi_spec



from typing import List, Optional

from starlette.applications import Starlette
from starlette.responses import JSONResponse, HTMLResponse
from starlette.routing import Route

from .schemas import build_schemas
from .route_def import RouteDef
from .auth import AuthBackend

def build_swagger_ui(title: str) -> str:
	return f'''
	<html>
	<head>
		<meta charset="UTF-8">
		<title>{title}</title>
		<link rel="stylesheet" type="text/css" href="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui.css">
	</head>
	<body>

	<div id="swagger-ui"></div>

	<script src="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui-standalone-preset.js"></script>
	<script src="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui-bundle.js"></script>

	<script>
	window.onload = function() {{
		const ui = SwaggerUIBundle({{
			url: "schema",
			dom_id: '#swagger-ui',
			deepLinking: true,
			presets: [
				SwaggerUIBundle.presets.apis,
				SwaggerUIStandalonePreset,
			],
			plugins: [
				SwaggerUIBundle.plugins.DownloadUrl,
			],
			layout: "BaseLayout",
		}})
		window.ui = ui
	}}
	</script>
	</body>
	</html>
	'''

def build_redoc_ui(title: str) -> str:
	return f'''
	<html>
	<head>
		<title>ReDoc - {title}</title>
		<meta charset="utf-8"/>
		<meta name="viewport" content="width=device-width, initial-scale=1">
		<link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
		</style>
	</head>
	<body>
		<redoc spec-url='schema'></redoc>
		<script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"> </script>
	</body>
	</html>
	'''

def openapi_app(
		title: str,
		route_defs: List[RouteDef],
		auth_backend: Optional[AuthBackend],
	) -> Starlette:
	swagger_ui = build_swagger_ui(title)
	redoc_ui = build_redoc_ui(title)

	openapi_schemas = build_schemas(route_defs, title, auth_backend)

	return Starlette(
		routes = [
			Route('/schema', lambda _: JSONResponse(openapi_schemas)),
			Route('/swagger', lambda _: HTMLResponse(swagger_ui)),
			Route('/redoc', lambda _: HTMLResponse(redoc_ui)),
		],
	)



from typing import Any, Union, Dict, Callable, Literal, Set, TypeVar
from dataclasses import dataclass
from datetime import datetime, date

from .errors import BadRequest
from .auth import Principal

_ParserType = Callable[[ str ], Any]

@dataclass
class ArgDef:
	original_type: type
	element_type: type
	optional: bool
	parser: _ParserType
	is_list: bool

def _parse_bool(x: str) -> bool:
	x = x.lower()
	if x == "true":
		return True
	elif x == "false":
		return False
	raise ValueError('expected true or false')

_parser_overrides: Dict[type, _ParserType] = {
	bool: _parse_bool,
	datetime: datetime.fromisoformat,
	date: date.fromisoformat,
}

T = TypeVar('T')
def create_enum_parser(options: Set[T], elem_parser: Callable[[ str ], T]) -> Callable[[ str ], T]:
	def _impl(x: str) -> T:
		if x in options:
			return elem_parser(x)
		else:
			raise ValueError(f'expected one of {options}')
	return _impl

def make_arg_def(original_type: type) -> ArgDef:
	element_type = original_type
	is_optional = getattr(element_type, '__origin__', None) == Union and getattr(element_type, '__args__')[1] == type(None)
	if is_optional:
		element_type = getattr(element_type, '__args__')[0]

	is_list = getattr(element_type, '__origin__', None) == list
	if is_list:
		element_type = getattr(element_type, '__args__')[0]

	parser = _parser_overrides.get(element_type, element_type)

	is_enum = getattr(element_type, '__origin__', None) == Literal
	if is_enum:
		enum_values = getattr(element_type, '__args__')
		element_type = type(enum_values[0])
		enum_values = set(enum_values)
		parser = create_enum_parser(enum_values, _parser_overrides.get(element_type, element_type))
	else:
		enum_values = set()
	
	return ArgDef(
		original_type = original_type,
		element_type = element_type,
		optional = is_optional,
		parser = parser,
		is_list = is_list,
	)

@dataclass
class RouteDef:
	path: str
	http_method: str
	impl: Callable[..., Any]
	args: Dict[str, ArgDef]
	return_type: type
	requires_principal: bool
	readable_name: str
	doc: str

def make_route_def(impl: Callable[..., Any]) -> RouteDef:
	name_tokens = impl.__name__.split('_')
	http_method = name_tokens[0]

	input_annotations = impl.__annotations__.copy()
	return_type = input_annotations['return']
	del input_annotations['return']
	requires_principal = input_annotations.get('principal') is not None
	if requires_principal:
		del input_annotations['principal']

	if http_method == 'post':
		if len(input_annotations) != 1:
			raise Exception('POST methods require one argument')

	args = {
		name: make_arg_def(element_type) for name, element_type
		in input_annotations.items()
	}

	return RouteDef(
		path = '/' + '_'.join(name_tokens[1:]),
		http_method = http_method,
		impl = impl,
		args = args,
		return_type = return_type,
		requires_principal = requires_principal,
		readable_name = ' '.join([ s.title() for s in name_tokens[1:] ]),
		doc = getattr(impl, '__doc__'),
	)



from setuptools import setup

with open('readme.md', 'r', encoding = 'utf-8') as fh:
	long_description = fh.read()

setup(
	name = 'govyn',
	version = '0.15.0',
	author = 'Ruan Pearce-Authers',
	author_email = 'ruanpa@outlook.com',
	description = 'HTTP APIs in typed Python',
	url = 'https://github.com/returnString/govyn',
	long_description = long_description,
	long_description_content_type = 'text/markdown',
	packages = [ 'govyn' ],
	package_data = {
		'govyn': [ 'py.typed' ],
	},
	zip_safe = False,
	install_requires = [
		'starlette >= 0.14',
		'uvicorn >= 0.13',
		'dacite >= 1.5.1',
		'aioprometheus[aiohttp] >= 20.0',
		# not explicit dependencies, but required to avoid build breaks
		'aiohttp >= 3.7',
	],
	classifiers=[
		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
	],
	python_requires = '>= 3.8',
)



from typing import Set, Optional, Dict, Protocol, Callable, TypeVar, Any, ClassVar, Tuple, cast
from dataclasses import dataclass
from abc import ABC, abstractmethod
from functools import wraps

from starlette.types import ASGIApp
from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

from .errors import Unauthorised, Forbidden
from .metrics import MetricsRegistry

@dataclass
class Principal:
	id: str
	privileges: Set[str]

class AuthBackend(Protocol):
	async def resolve_principal(self, req: Request) -> Optional[Principal]:
		...

	def openapi_spec(self) -> Tuple[str, Dict[str, Any]]:
		...

class AuthMiddleware(BaseHTTPMiddleware):
	def __init__(self, app: ASGIApp, auth_backend: AuthBackend, metrics_registry: MetricsRegistry) -> None:
		super().__init__(app)
		self.auth_backend = auth_backend
		self.principal_resolution_histogram = metrics_registry.histogram('api_auth_principal_resolution_seconds')

	async def dispatch(self, req: Request, call_next: RequestResponseEndpoint) -> Response:
		with self.principal_resolution_histogram.observe_time():
			principal = await self.auth_backend.resolve_principal(req)

		if principal is None:
			return Unauthorised('authentication failed').as_response()

		req.state.principal = principal
		return await call_next(req)

class HeaderAuthBackend(ABC):
	header: ClassVar[str]

	async def resolve_principal(self, req: Request) -> Optional[Principal]:
		token = req.headers.get(self.header)
		if not token:
			return None

		return await self.principal_from_header(token)

	def openapi_spec(self) -> Tuple[str, Dict[str, Any]]:
		return 'API key', {
			'type': 'apiKey',
			'in': 'header',
			'name': self.header,
		}

	@abstractmethod
	async def principal_from_header(self, value: str) -> Optional[Principal]:
		...

TFunc = TypeVar('TFunc', bound = Callable[..., Any])

def privileged(privilege: str) -> Callable[[ TFunc ], TFunc]:
	def _decorator(func: TFunc) -> TFunc:
		@wraps(func)
		async def _impl(*args: Any, **kwargs: Any) -> Any:
			principal = cast(Principal, kwargs['principal'])
			if privilege not in principal.privileges:
				raise Forbidden('insufficient privileges')

			return await func(*args, **kwargs)

		return cast(TFunc, _impl)
	return _decorator



from typing import Any, Optional, Dict, List, Protocol

from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Route, Mount
from starlette.middleware import Middleware
from starlette.types import ASGIApp

from .route_def import make_route_def
from .endpoint import make_endpoint
from .errors import JSONErrorMiddleware
from .auth import AuthBackend, AuthMiddleware
from .security import CORSConfig, permissive_cors_config, cors_middleware_from_config
from .metrics import MetricsMiddleware, MetricsRegistry
from .openapi import openapi_app

def create_app(
		srv: Any,
		name: Optional[str] = None,
		auth_backend: Optional[AuthBackend] = None,
		cors_config: Optional[CORSConfig] = None,
		metrics_port: Optional[int] = None,
	) -> ASGIApp:
	name = name or type(srv).__name__
	cors_config = cors_config or permissive_cors_config()

	http_methods = [ 'get', 'post' ]
	method_prefixes = tuple([ m + '_' for m in http_methods ])
	route_defs = [ make_route_def(getattr(srv, m)) for m in dir(srv) if m in http_methods or m.startswith(method_prefixes) ]

	metrics_registry = getattr(srv, 'metrics', None)
	if not metrics_registry or not isinstance(metrics_registry, MetricsRegistry):
		metrics_registry = MetricsRegistry()

	metrics_registry._const_labels['app'] = name

	async def metrics_async_init() -> None:
		if metrics_port:
			await metrics_registry._prom_svc.start(addr = '0.0.0.0', port = metrics_port)

	startup_funcs = [ metrics_async_init ]
	shutdown_funcs = []

	def _attach_lifecyle_methods(obj: Any) -> None:
		if startup_func := getattr(obj, 'startup', None):
			startup_funcs.append(startup_func)
		if shutdown_func := getattr(obj, 'shutdown', None):
			shutdown_funcs.append(shutdown_func)

	_attach_lifecyle_methods(srv)

	middleware = [ Middleware(JSONErrorMiddleware) ]
	if auth_backend:
		middleware.append(Middleware(AuthMiddleware, auth_backend = auth_backend, metrics_registry = metrics_registry))
		_attach_lifecyle_methods(auth_backend)

	core_app = Starlette(
		routes = [
			Route(r.path, make_endpoint(r), methods = [ r.http_method.upper() ])
			for r in route_defs
		],
		middleware = middleware,
	)

	health_app = Starlette(
		routes = [
			Route('/check', lambda _: JSONResponse({}))
		]
	)

	return Starlette(
		routes = [
			Mount('/openapi', openapi_app(name, route_defs, auth_backend)),
			Mount('/health', health_app),
			Mount('/', core_app),
		],
		on_startup = startup_funcs,
		on_shutdown = shutdown_funcs,
		middleware = [
			Middleware(MetricsMiddleware, metrics_registry = metrics_registry),
			cors_middleware_from_config(cors_config),
		],
	)



import json
from dataclasses import asdict, is_dataclass
from datetime import date, datetime
from enum import Enum, EnumMeta
from typing import Any, Awaitable, Callable, Dict

from dacite.config import Config
from dacite.core import from_dict
from dacite.exceptions import DaciteError
from starlette.requests import Request
from starlette.responses import Response

from .errors import BadRequest
from .route_def import ArgDef, RouteDef


def default_json_ser(obj: Any) -> Any:
	if isinstance(obj, (datetime, date)):
		return obj.isoformat()
	if isinstance(obj, Enum):
		return obj.value

	raise TypeError(f'type {type(obj)} is not serializable')

class GovynJSONResponse(Response):
	media_type = 'application/json'

	def render(self, content: Any) -> bytes:
		return json.dumps(
			content,
			ensure_ascii = False,
			allow_nan = False,
			indent = None,
			separators = (',', ':'),
			default = default_json_ser,
		).encode('utf-8')

def parse_value(arg: ArgDef, var_name: str, str_value: str) -> Any:
	try:
		return arg.parser(str_value)
	except ValueError as e:
		base_err = f'invalid value for field {var_name} of type {arg.element_type.__name__}:'
		if isinstance(arg.element_type, EnumMeta):
			raise BadRequest(f'{base_err} Must be one of {[e.value for e in arg.element_type]}') # type: ignore
		raise BadRequest(f'{base_err}: {str(e)}')

async def query_string_parser(req: Request, args: Dict[str, ArgDef]) -> Dict[str, Any]:
	ret: Dict[str, Any] = dict()
	for var_name, arg_def in args.items():
		if arg_def.is_list:
			ret[var_name] = [ parse_value(arg_def, var_name, v) for v in req.query_params.getlist(var_name) ]
		else:
			value = req.query_params.get(var_name)
			if value is None:
				if not arg_def.optional:
					raise BadRequest(f'missing required field: {var_name}')
				ret[var_name] = None
			else:
				ret[var_name] = parse_value(arg_def, var_name, value)
	return ret

async def json_body_parser(req: Request, args: Dict[str, ArgDef]) -> Dict[str, Any]:
	try:
		json_body = await req.json()
	except json.JSONDecodeError:
		raise BadRequest('Request body is not valid JSON')

	name = list(args)[0]
	arg_def = args[name]
	try:
		body: Any = from_dict(arg_def.element_type, json_body, Config(type_hooks = {
			datetime: datetime.fromisoformat,
			date: date.fromisoformat
		}, cast=[Enum]))
	except (DaciteError, ValueError) as e:
		raise BadRequest(str(e))

	return { name: body }

_parser_dict = {
	'get': query_string_parser,
	'post': json_body_parser,
}

def make_endpoint(route: RouteDef) -> Callable[[ Request ], Awaitable[Response]]:
	parser = _parser_dict[route.http_method]

	async def endpoint(req: Request) -> Response:
		args = await parser(req, route.args)
		if route.requires_principal:
			args['principal'] = req.state.principal
		res = await route.impl(**args)
		if is_dataclass(res):
			res = asdict(res)
		return GovynJSONResponse(res)

	return endpoint
